@@ -1,21 +1,27 @@
+from __future__ import annotations
+
+from collections.abc import Iterator
 from datetime import datetime, timezone
 from enum import Enum
 import logging
 import os
 import os.path
 from pathlib import Path
-from typing import IO, Dict, Iterator, Optional, Tuple, Union
+from types import SimpleNamespace, TracebackType
+from typing import IO, Any, Optional, Tuple, cast
 
+import aiohttp
+from aiohttp_retry import ListRetry, RetryClient
 from datalad.distribution.dataset import Dataset
 from datalad.support.annexrepo import AnnexRepo
 from datalad.utils import get_dataset_root
-import fsspec
 from fsspec.exceptions import BlocksizeMismatchError
 from fsspec.implementations.cached import CachingFileSystem
+from fsspec.implementations.http import HTTPFileSystem
 import methodtools
 
 from .consts import CACHE_SIZE
-from .utils import is_annex_dir_or_key
+from .utils import AnnexKey, is_annex_dir_or_key
 
 lgr = logging.getLogger("datalad.fuse.fsspec")
 
@@ -23,7 +29,7 @@ FileState = Enum("FileState", "NOT_ANNEXED NO_CONTENT HAS_CONTENT")
 
 
 class DatasetAdapter:
-    def __init__(self, path: Union[str, Path], mode_transparent: bool = False) -> None:
+    def __init__(self, path: str | Path, mode_transparent: bool = False) -> None:
         self.path = Path(path)
         self.mode_transparent = mode_transparent
         ds = Dataset(path)
@@ -36,7 +42,7 @@ class DatasetAdapter:
             ds.repo.get_commit_date(), tz=timezone.utc
         )
         self.fs = CachingFileSystem(
-            fs=fsspec.filesystem("http"),
+            fs=HTTPFileSystem(get_client=get_client),
             # target_protocol='blockcache',
             cache_storage=os.path.join(path, ".git", "datalad", "cache", "fsspec"),
             # cache_check=600,
@@ -51,19 +57,19 @@ class DatasetAdapter:
             self.annex._batched.clear()
 
     @methodtools.lru_cache(maxsize=CACHE_SIZE)
-    def get_file_state(self, relpath: str) -> Tuple[FileState, Optional[str]]:
+    def get_file_state(self, relpath: str) -> tuple[FileState, Optional[AnnexKey]]:
         p = self.path / relpath
         lgr.debug("get_file_state: %s", relpath)
 
-        def handle_path_under_annex_objects(p: Path):
+        def handle_path_under_annex_objects(
+            p: Path,
+        ) -> tuple[FileState, Optional[AnnexKey]]:
             iadok = is_annex_dir_or_key(p)
-            if iadok is not None and iadok[1] == "key":
-                assert iadok[0] == str(self.path)
-                key = filename2key(p.name)
+            if isinstance(iadok, AnnexKey):
                 if p.exists():
-                    return (FileState.HAS_CONTENT, key)
+                    return (FileState.HAS_CONTENT, iadok)
                 else:
-                    return (FileState.NO_CONTENT, key)
+                    return (FileState.NO_CONTENT, iadok)
             else:
                 return (FileState.NOT_ANNEXED, None)
 
@@ -75,7 +81,7 @@ class DatasetAdapter:
         if not p.is_symlink():
             if p.stat().st_size < 1024 and self.annex is not None:
                 if self.annex.is_under_annex(relpath, batch=True):
-                    key = self.annex.get_file_key(relpath, batch=True)
+                    key = AnnexKey.parse(self.annex.get_file_key(relpath, batch=True))
                     if self.annex.file_has_content(relpath, batch=True):
                         return (FileState.HAS_CONTENT, key)
                     else:
@@ -163,10 +169,10 @@ class DatasetAdapter:
             )
         if fstate is FileState.NO_CONTENT:
             lgr.debug("%s: opening via fsspec", relpath)
-            for url in self.get_urls(key):
+            for url in self.get_urls(str(key)):
                 try:
                     lgr.debug("%s: Attempting to open via URL %s", relpath, url)
-                    return self.fs.open(url, mode, **kwargs)
+                    return self.fs.open(url, mode, **kwargs)  # type: ignore
                 except BlocksizeMismatchError as e:
                     lgr.warning(
                         "%s: Blocksize mismatch: %s; deleting cached file and"
@@ -175,30 +181,37 @@ class DatasetAdapter:
                         e,
                     )
                     self.fs.pop_from_cache(url)
-                    return self.fs.open(url, mode, **kwargs)
+                    return self.fs.open(url, mode, **kwargs)  # type: ignore
                 except FileNotFoundError as e:
                     lgr.debug(
                         "Failed to open file %s at URL %s: %s", relpath, url, str(e)
                     )
-            raise IOError(f"Could not find a usable URL for {relpath}")
+            raise IOError(
+                f"Could not find a usable URL for {relpath} within {self.path}"
+            )
         else:
             lgr.debug("%s: opening directly", relpath)
-            return open(self.path / relpath, mode, **kwargs)
+            return open(self.path / relpath, mode, **kwargs)  # type: ignore
 
     def clear(self) -> None:
         self.fs.clear_cache()
 
 
 class FsspecAdapter:
-    def __init__(self, root: Union[str, Path], mode_transparent: bool = False) -> None:
+    def __init__(self, root: str | Path, mode_transparent: bool = False) -> None:
         self.root = Path(root)
         self.mode_transparent = mode_transparent
-        self.datasets: Dict[Path, DatasetAdapter] = {}
+        self.datasets: dict[Path, DatasetAdapter] = {}
 
-    def __enter__(self):
+    def __enter__(self) -> FsspecAdapter:
         return self
 
-    def __exit__(self, _exc_type, _exc_val, _exc_tb):
+    def __exit__(
+        self,
+        _exc_type: Optional[type[BaseException]],
+        _exc_val: Optional[BaseException],
+        _exc_tb: Optional[TracebackType],
+    ) -> None:
         for ds in self.datasets.values():
             ds.close()
         self.datasets.clear()
@@ -206,19 +219,20 @@ class FsspecAdapter:
     @methodtools.lru_cache(maxsize=CACHE_SIZE)
     # TODO: optimize "caching" more since for all files under the same directory
     # they all would belong to the same dataset
-    def get_dataset_path(self, path: Union[str, Path]) -> Path:
+    def get_dataset_path(self, path: str | Path) -> Path:
         path = Path(self.root, path)
         dspath = get_dataset_root(path)
         if dspath is None:
             raise ValueError(f"Path not under DataLad: {path}")
         dspath = Path(dspath)
+        assert isinstance(dspath, Path)
         try:
             dspath.relative_to(self.root)
         except ValueError:
             raise ValueError(f"Path not under root dataset: {path}")
         return dspath
 
-    def resolve_dataset(self, filepath: Union[str, Path]) -> Tuple[DatasetAdapter, str]:
+    def resolve_dataset(self, filepath: str | Path) -> tuple[DatasetAdapter, str]:
         dspath = self.get_dataset_path(filepath)
         try:
             dsap = self.datasets[dspath]
@@ -231,7 +245,7 @@ class FsspecAdapter:
 
     def open(
         self,
-        filepath: Union[str, Path],
+        filepath: str | Path,
         mode: str = "rb",
         encoding: str = "utf-8",
         errors: Optional[str] = None,
@@ -242,12 +256,18 @@ class FsspecAdapter:
         )
         return dsap.open(relpath, mode=mode, encoding=encoding, errors=errors)
 
-    def is_under_annex(self, filepath: Union[str, Path]) -> bool:
+    def get_file_state(
+        self, filepath: str | Path
+    ) -> tuple[FileState, Optional[AnnexKey]]:
+        dsap, relpath = self.resolve_dataset(filepath)
+        return cast(Tuple[FileState, Optional[AnnexKey]], dsap.get_file_state(relpath))
+
+    def is_under_annex(self, filepath: str | Path) -> bool:
         dsap, relpath = self.resolve_dataset(filepath)
         fstate, _ = dsap.get_file_state(relpath)
         return fstate is not FileState.NOT_ANNEXED
 
-    def get_commit_datetime(self, filepath: Union[str, Path]) -> datetime:
+    def get_commit_datetime(self, filepath: str | Path) -> datetime:
         dsap, _ = self.resolve_dataset(filepath)
         return dsap.commit_dt
 
@@ -256,9 +276,22 @@ def is_http_url(s: str) -> bool:
     return s.lower().startswith(("http://", "https://"))
 
 
-def filename2key(name: str) -> str:
-    # See `keyFile` and `fileKey` in `Annex/Locations.hs` in the git-annex
-    # source
-    return (
-        name.replace("%", "/").replace("&c", ":").replace("&s", "%").replace("&a", "&")
+async def on_request_start(
+    _session: aiohttp.ClientSession,
+    trace_config_ctx: SimpleNamespace,
+    params: aiohttp.TraceRequestStartParams,
+) -> None:
+    if trace_config_ctx.trace_request_ctx["current_attempt"] > 1:
+        lgr.warning("Retrying request to %s", params.url)
+
+
+async def get_client(**kwargs: Any) -> RetryClient:
+    trace_config = aiohttp.TraceConfig()
+    trace_config.on_request_start.append(on_request_start)
+    return RetryClient(
+        client_session=aiohttp.ClientSession(
+            trace_configs=[trace_config],
+            **kwargs,
+        ),
+        retry_options=ListRetry(timeouts=[1, 2, 6, 15, 36]),
     )
